# univi/utils/io.py

from __future__ import annotations

from typing import Any, Dict, Optional, Sequence, Union, Mapping

import os
import json

import numpy as np
import torch
import scipy.sparse as sp
import anndata as ad


def save_checkpoint(
    path: str,
    model_state: Dict[str, Any],
    optimizer_state: Optional[Dict[str, Any]] = None,
    extra: Optional[Dict[str, Any]] = None,
) -> None:
    payload: Dict[str, Any] = {"model_state": model_state}
    if optimizer_state is not None:
        payload["optimizer_state"] = optimizer_state
    if extra is not None:
        payload["extra"] = extra
    torch.save(payload, path)


def load_checkpoint(path: str) -> Dict[str, Any]:
    return torch.load(path, map_location="cpu")


def save_config_json(config: Dict[str, Any], path: str) -> None:
    with open(path, "w") as f:
        json.dump(config, f, indent=2, sort_keys=True)


def load_config(config_path: str) -> Dict[str, Any]:
    with open(config_path) as f:
        cfg = json.load(f)

    cfg.setdefault("model", {})
    model_cfg = cfg["model"]

    # Backwards-compat defaults
    model_cfg.setdefault("loss_mode", "v1")
    model_cfg.setdefault("v1_recon", "cross")
    model_cfg.setdefault("v1_recon_mix", 0.0)
    model_cfg.setdefault("normalize_v1_terms", True)

    # Old configs used a single dropout
    base_do = model_cfg.get("dropout", 0.0)
    model_cfg.setdefault("encoder_dropout", base_do)
    model_cfg.setdefault("decoder_dropout", base_do)

    model_cfg.setdefault("beta", 5.0)
    model_cfg.setdefault("gamma", 60.0)

    if "training" in cfg:
        cfg["training"].setdefault("device", "cuda")

    return cfg


def save_anndata_splits(
    adata: ad.AnnData,
    outdir: str,
    prefix: str = "dataset",
    split_key: Optional[str] = "split",
    split_map: Optional[Dict[str, Union[Sequence[int], Sequence[str]]]] = None,
    train_label: str = "train",
    val_label: str = "val",
    test_label: str = "test",
    copy: bool = True,
    write_backed: bool = False,
) -> Dict[str, ad.AnnData]:
    os.makedirs(outdir, exist_ok=True)

    def subset_from_idx(idx_list: Sequence[Union[int, str]]) -> ad.AnnData:
        if len(idx_list) == 0:
            return adata[:0].copy() if copy else adata[:0]
        first = idx_list[0]
        if isinstance(first, str):
            return adata[list(idx_list)]
        return adata[list(idx_list), :]

    if split_map is not None:
        train = subset_from_idx(list(split_map.get("train", [])))
        val = subset_from_idx(list(split_map.get("val", [])))
        test = subset_from_idx(list(split_map.get("test", [])))
    else:
        if split_key is None or split_key not in adata.obs:
            raise ValueError(
                "Expected split labels in adata.obs[%r], or provide split_map={'train':...,'val':...,'test':...}."
                % split_key
            )
        s = adata.obs[split_key].astype(str)
        train = adata[s == train_label]
        val = adata[s == val_label]
        test = adata[s == test_label]

    if copy:
        train = train.copy()
        val = val.copy()
        test = test.copy()

    paths = {
        "train": os.path.join(outdir, "%s_train.h5ad" % prefix),
        "val": os.path.join(outdir, "%s_val.h5ad" % prefix),
        "test": os.path.join(outdir, "%s_test.h5ad" % prefix),
    }

    train.write_h5ad(paths["train"])
    val.write_h5ad(paths["val"])
    test.write_h5ad(paths["test"])

    return {"train": train, "val": val, "test": test}


def _select_X(
    adata_obj: ad.AnnData,
    layer: Optional[str],
    X_key: str,
):
    # X_key != "X" => obsm
    if X_key != "X":
        if X_key not in adata_obj.obsm:
            raise KeyError("X_key=%r not found in adata.obsm. Keys=%s" % (X_key, list(adata_obj.obsm.keys())))
        return adata_obj.obsm[X_key]

    # X_key == "X" => layers[layer] or X
    if layer is not None:
        if layer not in adata_obj.layers:
            raise KeyError("layer=%r not found in adata.layers. Keys=%s" % (layer, list(adata_obj.layers.keys())))
        return adata_obj.layers[layer]
    return adata_obj.X


@torch.no_grad()
def write_univi_latent(
    model,
    adata_dict: Dict[str, ad.AnnData],
    *,
    obsm_key: str = "X_univi",
    batch_size: int = 512,
    device: Optional[Union[str, torch.device]] = None,
    use_mean: bool = False,
    # NEW: support layer / X_key (global or per-modality)
    layer: Union[None, str, Mapping[str, Optional[str]]] = None,
    X_key: Union[str, Mapping[str, str]] = "X",
    require_paired: bool = True,
) -> np.ndarray:
    """
    Encode paired modalities and write a *shared/fused* latent Z to each adata.obsm[obsm_key].

    Notes
    -----
    - This writes THE SAME Z into every modality AnnData (because it's fused).
    - If you want modality-specific means (mu_modality), use evaluation.encode_adata(..., latent="modality_mean").
    """
    model.eval()

    names = list(adata_dict.keys())
    if len(names) == 0:
        raise ValueError("adata_dict is empty.")

    n = int(adata_dict[names[0]].n_obs)

    if require_paired:
        ref = adata_dict[names[0]].obs_names.values
        for nm in names[1:]:
            if adata_dict[nm].n_obs != n:
                raise ValueError("n_obs mismatch: %s has %d vs %d" % (nm, adata_dict[nm].n_obs, n))
            if not np.array_equal(adata_dict[nm].obs_names.values, ref):
                raise ValueError("obs_names mismatch between %s and %s" % (names[0], nm))

    # device default: model device
    if device is None:
        try:
            device = next(model.parameters()).device
        except StopIteration:
            device = "cpu"

    # normalize layer/X_key to per-modality mappings
    if isinstance(layer, dict):
        layer_by_mod = dict(layer)
    else:
        layer_by_mod = {nm: layer for nm in names}

    if isinstance(X_key, dict):
        xkey_by_mod = dict(X_key)
    else:
        xkey_by_mod = {nm: X_key for nm in names}

    zs = []
    for start in range(0, n, int(batch_size)):
        end = min(n, start + int(batch_size))
        x_dict = {}

        for nm in names:
            adata_obj = adata_dict[nm]
            X = _select_X(adata_obj, layer_by_mod.get(nm, None), xkey_by_mod.get(nm, "X"))
            xb = X[start:end]
            if sp.issparse(xb):
                xb = xb.toarray()
            else:
                xb = np.asarray(xb)
            x_dict[nm] = torch.as_tensor(xb, dtype=torch.float32, device=device)

        # forward compat: some forwards accept epoch or y; safest is plain model(x_dict)
        out = model(x_dict)
        z = out["mu_z"] if (use_mean and ("mu_z" in out)) else out["z"]
        zs.append(z.detach().cpu().numpy())

    Z = np.vstack(zs)

    for nm in names:
        adata_dict[nm].obsm[obsm_key] = Z

    return Z




# univi/__init__.py

from __future__ import annotations

from typing import Any, List

__version__ = "0.2.2"

from .config import ModalityConfig, UniVIConfig, TrainingConfig
from .models import UniVIMultiModalVAE
from . import matching

# Public API that is safe to import eagerly:
__all__ = [
    "__version__",
    "ModalityConfig",
    "UniVIConfig",
    "TrainingConfig",
    "UniVIMultiModalVAE",
    "matching",
    # Lazy exports (resolved via __getattr__)
    "UniVITrainer",
    "write_univi_latent",
    "pipeline",
    "diagnostics",
]


def __getattr__(name: str) -> Any:
    """
    Lazy exports keep `import univi` fast/light and avoid heavy deps unless needed.
    """
    if name == "UniVITrainer":
        from .trainer import UniVITrainer
        return UniVITrainer

    if name == "write_univi_latent":
        from .utils.io import write_univi_latent
        return write_univi_latent

    # Note: these return the *module* objects
    if name == "pipeline":
        from . import pipeline as _pipeline
        return _pipeline

    if name == "diagnostics":
        from . import diagnostics as _diagnostics
        return _diagnostics

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__() -> List[str]:
    # Makes autocomplete / dir(univi) show lazy symbols too.
    return sorted(list(globals().keys()) + __all__)




# univi/data.py

from __future__ import annotations

from typing import Any, Dict, Mapping, Optional, Tuple, Union, List

import os
import numpy as np
import scipy.sparse as sp
import anndata as ad
from anndata import AnnData

import torch
from torch.utils.data import Dataset


LayerSpec = Union[None, str, Mapping[str, Optional[str]]]
XKeySpec = Union[str, Mapping[str, str]]


def _get_matrix(adata_obj: AnnData, *, layer: Optional[str], X_key: str):
    if X_key != "X":
        if X_key not in adata_obj.obsm:
            raise KeyError("X_key=%r not found in adata.obsm. Keys=%s" % (X_key, list(adata_obj.obsm.keys())))
        return adata_obj.obsm[X_key]
    if layer is not None:
        if layer not in adata_obj.layers:
            raise KeyError("layer=%r not found in adata.layers. Keys=%s" % (layer, list(adata_obj.layers.keys())))
        return adata_obj.layers[layer]
    return adata_obj.X


def infer_input_dim(adata_obj: AnnData, *, layer: Optional[str], X_key: str) -> int:
    X = _get_matrix(adata_obj, layer=layer, X_key=X_key)
    if not hasattr(X, "shape") or len(X.shape) != 2:
        raise ValueError("Selected matrix for (layer=%r, X_key=%r) is not 2D." % (layer, X_key))
    return int(X.shape[1])


def align_paired_obs_names(
    adata_dict: Dict[str, AnnData],
    how: str = "intersection",
    require_nonempty: bool = True,
) -> Dict[str, AnnData]:
    if not adata_dict:
        raise ValueError("adata_dict is empty")
    if how != "intersection":
        raise ValueError("Unsupported how=%r. Only 'intersection' is supported." % how)

    names = list(adata_dict.keys())
    shared = None
    for nm in names:
        idx = adata_dict[nm].obs_names
        shared = idx if shared is None else shared.intersection(idx)

    if require_nonempty and (shared is None or len(shared) == 0):
        raise ValueError("No shared obs_names across modalities (intersection is empty).")

    shared = shared.sort_values()
    out: Dict[str, AnnData] = {}
    for nm in names:
        out[nm] = adata_dict[nm].loc[shared].copy()
    return out


class MultiModalDataset(Dataset):
    """
    Returns either:
      - x_dict
      - (x_dict, y) if labels are provided
    """

    def __init__(
        self,
        adata_dict: Dict[str, AnnData],
        layer: LayerSpec = None,
        X_key: XKeySpec = "X",
        paired: bool = True,
        device: Optional[torch.device] = None,
        labels: Optional[Union[np.ndarray, torch.Tensor, List[int]]] = None,
    ):
        self.adata_dict = adata_dict
        self.paired = bool(paired)
        self.device = device
        self.labels = labels

        if isinstance(layer, Mapping):
            self.layer_by_modality = dict(layer)
        else:
            self.layer_by_modality = {k: layer for k in adata_dict.keys()}

        if isinstance(X_key, Mapping):
            self.xkey_by_modality = dict(X_key)
        else:
            self.xkey_by_modality = {k: X_key for k in adata_dict.keys()}

        if self.paired:
            first = next(iter(adata_dict.values()))
            for nm, adata_obj in self.adata_dict.items():
                if adata_obj.n_obs != first.n_obs:
                    raise ValueError("Paired dataset requires matching n_obs across modalities.")
                if not np.array_equal(adata_obj.obs_names.values, first.obs_names.values):
                    raise ValueError("Paired dataset requires identical obs_names order; %r differs." % nm)

        if self.labels is not None:
            if hasattr(self.labels, "__len__") and len(self.labels) != self.__len__():
                raise ValueError("labels length (%d) must equal n_cells (%d)" % (len(self.labels), self.__len__()))

    def __len__(self) -> int:
        first = next(iter(self.adata_dict.values()))
        return int(first.n_obs)

    def _get_X_row(self, adata_obj: AnnData, idx: int, layer: Optional[str], X_key: str) -> np.ndarray:
        X = _get_matrix(adata_obj, layer=layer, X_key=X_key)
        row = X[idx]
        if sp.issparse(row):
            row = row.toarray()
        row = np.asarray(row).reshape(-1).astype(np.float32)
        return row

    def __getitem__(self, idx: int):
        out: Dict[str, torch.Tensor] = {}
        for name, adata_obj in self.adata_dict.items():
            layer = self.layer_by_modality.get(name, None)
            xkey = self.xkey_by_modality.get(name, "X")
            row_np = self._get_X_row(adata_obj, idx, layer=layer, X_key=xkey)
            t = torch.from_numpy(row_np)
            if self.device is not None:
                t = t.to(self.device)
            out[name] = t

        if self.labels is None:
            return out

        y = self.labels[idx]
        if not torch.is_tensor(y):
            y = torch.as_tensor(y, dtype=torch.long)
        else:
            y = y.long()
        if self.device is not None:
            y = y.to(self.device)

        return out, y


def dataset_from_anndata_dict(
    adata_dict: Dict[str, AnnData],
    layer: LayerSpec = None,
    X_key: XKeySpec = "X",
    paired: bool = True,
    align_obs: bool = True,
    labels: Optional[Union[np.ndarray, torch.Tensor, List[int]]] = None,
) -> Tuple[MultiModalDataset, Dict[str, AnnData]]:
    if align_obs and paired:
        adata_dict = align_paired_obs_names(adata_dict, how="intersection")
    ds = MultiModalDataset(adata_dict, layer=layer, X_key=X_key, paired=paired, labels=labels)
    return ds, adata_dict


def load_anndata_dict_from_config(
    modality_cfgs: List[Dict[str, Any]],
    data_root: Optional[str] = None,
) -> Dict[str, AnnData]:
    out: Dict[str, AnnData] = {}
    for m in modality_cfgs:
        name = m["name"]
        path = m["h5ad_path"]
        if data_root is not None and not os.path.isabs(path):
            path = os.path.join(data_root, path)
        out[name] = ad.read_h5ad(path)
    return out




# univi/trainer.py

from __future__ import annotations

from dataclasses import asdict
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from torch import nn, optim
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
import scipy.sparse as sp

from .config import TrainingConfig
from .utils.logging import get_logger

BatchType = Union[Dict[str, torch.Tensor], Tuple[Dict[str, torch.Tensor], torch.Tensor]]


class UniVITrainer:
    def __init__(
        self,
        model: nn.Module,
        train_loader: DataLoader,
        val_loader: Optional[DataLoader] = None,
        train_cfg: Optional[TrainingConfig] = None,
        device: Optional[str] = None,
    ):
        self.model = model
        self.train_loader = train_loader
        self.val_loader = val_loader

        self.cfg = train_cfg or TrainingConfig()
        self.device = device or self.cfg.device

        self.model.to(self.device)
        self.optimizer = optim.Adam(
            self.model.parameters(),
            lr=float(self.cfg.lr),
            weight_decay=float(self.cfg.weight_decay),
        )

        self.logger = get_logger("UniVITrainer")

        self.best_val_loss = float("inf")
        self.best_state_dict: Optional[Dict[str, Any]] = None
        self.epochs_no_improve = 0
        self.best_epoch: Optional[int] = None

        self.history: Dict[str, List[float]] = {
            "train_loss": [],
            "val_loss": [],
            "beta": [],
            "gamma": [],
        }

        self._log_config()

    def fit(self) -> Dict[str, List[float]]:
        epoch_iter = tqdm(range(1, int(self.cfg.n_epochs) + 1), desc="Training UniVI", leave=True)

        for epoch in epoch_iter:
            tr_loss, tr_beta, tr_gamma = self._run_one_epoch(epoch, train=True)
            self.history["train_loss"].append(tr_loss)
            self.history["beta"].append(tr_beta)
            self.history["gamma"].append(tr_gamma)

            if self.val_loader is not None:
                va_loss, _, _ = self._run_one_epoch(epoch, train=False)
                self.history["val_loss"].append(va_loss)

                epoch_iter.set_postfix(
                    train_loss="%.4f" % tr_loss,
                    val_loss="%.4f" % va_loss,
                    beta="%.3f" % tr_beta,
                    gamma="%.3f" % tr_gamma,
                )

                self._maybe_early_stop(va_loss, epoch)
                if self._should_stop():
                    self.logger.info("Early stopping at epoch %d (best val loss=%.4f)" % (epoch, self.best_val_loss))
                    break
            else:
                epoch_iter.set_postfix(
                    train_loss="%.4f" % tr_loss,
                    beta="%.3f" % tr_beta,
                    gamma="%.3f" % tr_gamma,
                )

        if self.val_loader is not None and self.best_state_dict is not None:
            self.model.load_state_dict(self.best_state_dict)
            if self.best_epoch is not None:
                self.logger.info("Restored best model from epoch %d (val loss=%.4f)" % (self.best_epoch, self.best_val_loss))

        return self.history

    def _split_batch(self, batch: BatchType) -> Tuple[Dict[str, torch.Tensor], Optional[torch.Tensor]]:
        if isinstance(batch, (tuple, list)) and len(batch) == 2:
            x_dict, y = batch
        else:
            x_dict, y = batch, None

        if not isinstance(x_dict, dict):
            raise TypeError("Expected batch to be dict or (dict, y). Got %r" % type(x_dict))

        x_out: Dict[str, torch.Tensor] = {}
        for k, v in x_dict.items():
            if v is None:
                x_out[k] = None
            elif torch.is_tensor(v):
                x_out[k] = v.to(self.device)
            else:
                x_out[k] = torch.as_tensor(v, dtype=torch.float32, device=self.device)

        if y is not None:
            if not torch.is_tensor(y):
                y = torch.as_tensor(y, dtype=torch.long)
            y = y.long().to(self.device)

        return x_out, y

    def _as_float(self, v: Any) -> float:
        if v is None:
            return 0.0
        if isinstance(v, (float, int)):
            return float(v)
        if torch.is_tensor(v):
            return float(v.detach().cpu().item())
        try:
            return float(v)
        except Exception:
            return 0.0

    def _forward_model(self, x_dict: Dict[str, torch.Tensor], epoch: int, y: Optional[torch.Tensor]):
        # Prefer newest signature
        try:
            return self.model(x_dict, epoch=epoch, y=y)
        except TypeError:
            pass
        # epoch only
        try:
            return self.model(x_dict, epoch=epoch)
        except TypeError:
            pass
        # y only
        if y is not None:
            try:
                return self.model(x_dict, y=y)
            except TypeError:
                pass
        # oldest
        return self.model(x_dict)

    def _run_one_epoch(self, epoch: int, train: bool = True) -> Tuple[float, float, float]:
        if train:
            self.model.train()
            loader = self.train_loader
        else:
            self.model.eval()
            if self.val_loader is None:
                raise ValueError("val_loader is None but train=False")
            loader = self.val_loader

        total_loss = 0.0
        total_beta = 0.0
        total_gamma = 0.0
        n_batches = 0

        for batch in loader:
            x_dict, y = self._split_batch(batch)

            if train:
                self.optimizer.zero_grad(set_to_none=True)

            with torch.set_grad_enabled(train):
                out = self._forward_model(x_dict, epoch=epoch, y=y)
                loss = out["loss"]

                if train:
                    loss.backward()
                    if self.cfg.grad_clip is not None and float(self.cfg.grad_clip) > 0:
                        nn.utils.clip_grad_norm_(self.model.parameters(), float(self.cfg.grad_clip))
                    self.optimizer.step()

            total_loss += float(loss.detach().cpu().item())
            total_beta += self._as_float(out.get("beta", 0.0))
            total_gamma += self._as_float(out.get("gamma", 0.0))
            n_batches += 1

        avg_loss = total_loss / max(1, n_batches)
        avg_beta = total_beta / max(1, n_batches)
        avg_gamma = total_gamma / max(1, n_batches)

        if epoch % int(self.cfg.log_every) == 0 or epoch == 1:
            self.logger.info("[Epoch %03d] %s loss=%.4f (beta=%.3f, gamma=%.3f)" %
                             (epoch, "Train" if train else "Val", avg_loss, avg_beta, avg_gamma))

        return avg_loss, avg_beta, avg_gamma

    def _maybe_early_stop(self, val_loss: float, epoch: int) -> None:
        if not self.cfg.early_stopping:
            return

        improved = (self.best_val_loss - float(val_loss)) > float(self.cfg.min_delta)
        if improved:
            self.best_val_loss = float(val_loss)
            self.best_state_dict = {k: v.detach().cpu().clone() for k, v in self.model.state_dict().items()}
            self.best_epoch = int(epoch)
            self.epochs_no_improve = 0
            self.logger.info("[Epoch %03d] New best val loss: %.4f" % (epoch, val_loss))
        else:
            self.epochs_no_improve += 1

    def _should_stop(self) -> bool:
        if not self.cfg.early_stopping:
            return False
        return int(self.epochs_no_improve) >= int(self.cfg.patience)

    def encode_modality(
        self,
        adata,
        modality: str,
        layer: Optional[str] = None,
        X_key: str = "X",
        batch_size: int = 512,
    ) -> np.ndarray:
        names = getattr(self.model, "modality_names", None)
        if names is not None and modality not in names:
            raise ValueError("Unknown modality %r. Available: %s" % (modality, names))

        self.model.eval()
        X = adata.layers[layer] if layer is not None else adata.X
        if sp.issparse(X):
            X = X.toarray()
        X = np.asarray(X, dtype=np.float32)

        zs = []
        with torch.no_grad():
            for start in range(0, X.shape[0], int(batch_size)):
                xb = torch.as_tensor(X[start:start + int(batch_size)], dtype=torch.float32, device=self.device)
                mu_dict, logvar_dict = self.model.encode_modalities({modality: xb})
                mu_z, _ = self.model.mixture_of_experts(mu_dict, logvar_dict)
                zs.append(mu_z.detach().cpu().numpy())

        return np.vstack(zs)

    def _log_config(self) -> None:
        cfg_dict = asdict(self.cfg)
        self.logger.info("TrainingConfig:")
        for k, v in cfg_dict.items():
            self.logger.info("  %s: %r" % (k, v))




# univi/evaluation.py

from __future__ import annotations

from typing import Any, Dict, Optional, Tuple

import numpy as np
import scipy.sparse as sp
import torch

from sklearn.neighbors import NearestNeighbors
from sklearn.metrics import accuracy_score, confusion_matrix


def compute_foscttm(Z1: np.ndarray, Z2: np.ndarray, metric: str = "euclidean") -> float:
    Z1 = np.asarray(Z1)
    Z2 = np.asarray(Z2)
    if Z1.shape != Z2.shape:
        raise ValueError("Z1/Z2 must have same shape for FOSCTTM. Got %r vs %r" % (Z1.shape, Z2.shape))

    n = int(Z1.shape[0])
    if n <= 1:
        return 0.0

    nn = NearestNeighbors(n_neighbors=n, metric=metric)
    nn.fit(Z2)
    _, idx = nn.kneighbors(Z1, return_distance=True)

    ranks = np.empty(n, dtype=float)
    for i in range(n):
        pos = np.where(idx[i] == i)[0]
        ranks[i] = (int(pos[0]) / (n - 1)) if pos.size else 1.0

    return float(ranks.mean())


def compute_modality_mixing(
    Z: np.ndarray,
    modality_labels: np.ndarray,
    k: int = 20,
    metric: str = "euclidean",
) -> float:
    Z = np.asarray(Z)
    modality_labels = np.asarray(modality_labels)
    if Z.shape[0] != modality_labels.shape[0]:
        raise ValueError("Z and modality_labels must align on n_cells.")

    n = int(Z.shape[0])
    if n <= 1:
        return 0.0

    k_eff = int(min(max(int(k), 1), n - 1))
    nn = NearestNeighbors(n_neighbors=k_eff + 1, metric=metric)
    nn.fit(Z)
    _, neigh_idx = nn.kneighbors(Z)

    neigh_idx = neigh_idx[:, 1:]
    neigh_mods = modality_labels[neigh_idx]
    frac_other = (neigh_mods != modality_labels[:, None]).mean(axis=1)
    return float(frac_other.mean())


def label_transfer_knn(
    Z_source: np.ndarray,
    labels_source: np.ndarray,
    Z_target: np.ndarray,
    labels_target: Optional[np.ndarray] = None,
    k: int = 15,
    metric: str = "euclidean",
    return_label_order: bool = False,  # <-- NEW (default keeps your old scripts working)
) -> Tuple[np.ndarray, Optional[float], np.ndarray]:
    """
    Majority-vote kNN label transfer from source → target.

    Default return (backwards compatible):
        pred_labels, accuracy (or None), confusion_matrix (or empty)

    If return_label_order=True and labels_target is not None:
        pred_labels, accuracy, confusion_matrix, label_order
    """
    Z_source = np.asarray(Z_source)
    Z_target = np.asarray(Z_target)
    labels_source = np.asarray(labels_source)

    if labels_target is not None:
        labels_target = np.asarray(labels_target)

    n_source = int(Z_source.shape[0])
    if n_source == 0:
        raise ValueError("Z_source is empty.")

    k_eff = int(min(max(int(k), 1), n_source))
    nn = NearestNeighbors(n_neighbors=k_eff, metric=metric)
    nn.fit(Z_source)
    _, neigh_idx = nn.kneighbors(Z_target)

    # Fast majority vote that supports string labels:
    # map labels_source -> integer codes
    uniq_src, src_codes = np.unique(labels_source, return_inverse=True)

    pred_codes = np.empty(Z_target.shape[0], dtype=np.int64)
    for i in range(Z_target.shape[0]):
        votes = src_codes[neigh_idx[i]]
        # bincount on codes
        bc = np.bincount(votes, minlength=len(uniq_src))
        pred_codes[i] = int(bc.argmax())

    pred_labels = uniq_src[pred_codes]

    if labels_target is None:
        empty = np.array([])
        return pred_labels, None, empty

    label_order = np.unique(np.concatenate([labels_target, pred_labels]))
    acc = float(accuracy_score(labels_target, pred_labels))
    cm = confusion_matrix(labels_target, pred_labels, labels=label_order)

    if return_label_order:
        return pred_labels, acc, cm, label_order  # type: ignore[return-value]

    return pred_labels, acc, cm


def mse_per_feature(x_true: np.ndarray, x_pred: np.ndarray) -> np.ndarray:
    x_true = np.asarray(x_true)
    x_pred = np.asarray(x_pred)
    if x_true.shape != x_pred.shape:
        raise ValueError("x_true and x_pred must have same shape.")
    return np.mean((x_true - x_pred) ** 2, axis=0)


def pearson_corr_per_feature(x_true: np.ndarray, x_pred: np.ndarray) -> np.ndarray:
    x_true = np.asarray(x_true)
    x_pred = np.asarray(x_pred)
    if x_true.shape != x_pred.shape:
        raise ValueError("x_true and x_pred must have same shape.")

    x_true_c = x_true - x_true.mean(axis=0, keepdims=True)
    x_pred_c = x_pred - x_pred.mean(axis=0, keepdims=True)

    num = (x_true_c * x_pred_c).sum(axis=0)
    denom = np.sqrt((x_true_c ** 2).sum(axis=0) * (x_pred_c ** 2).sum(axis=0)) + 1e-8
    return num / denom


def reconstruction_metrics(x_true: np.ndarray, x_pred: np.ndarray) -> Dict[str, Any]:
    pf_mse = mse_per_feature(x_true, x_pred)
    pf_r = pearson_corr_per_feature(x_true, x_pred)
    return {
        "mse_mean": float(np.mean(pf_mse)),
        "mse_median": float(np.median(pf_mse)),
        "pearson_mean": float(np.mean(pf_r)),
        "pearson_median": float(np.median(pf_r)),
        "mse_per_feature": pf_mse,
        "pearson_per_feature": pf_r,
    }


def encode_adata(
    model,
    adata,
    modality: str,
    device: str = "cpu",
    layer: Optional[str] = None,
    X_key: str = "X",
    batch_size: int = 1024,
    latent: str = "moe_mean",
    random_state: int = 0,
) -> np.ndarray:
    from .data import _get_matrix

    latent = str(latent).lower().strip()
    valid = {"moe_mean", "moe_sample", "modality_mean", "modality_sample"}
    if latent not in valid:
        raise ValueError("latent must be one of %s; got %r" % (sorted(valid), latent))

    def _sample_gaussian(mu: torch.Tensor, logvar: torch.Tensor, gen: torch.Generator) -> torch.Tensor:
        eps = torch.randn(mu.shape, device=mu.device, generator=gen, dtype=mu.dtype)
        return mu + eps * torch.exp(0.5 * logvar)

    model.eval()
    X = _get_matrix(adata, layer=layer, X_key=X_key)
    if sp.issparse(X):
        X = X.toarray()

    gen = torch.Generator(device=device)
    gen.manual_seed(int(random_state))

    zs = []
    with torch.no_grad():
        for start in range(0, X.shape[0], int(batch_size)):
            end = min(start + int(batch_size), X.shape[0])
            xb = torch.as_tensor(np.asarray(X[start:end]), dtype=torch.float32, device=device)

            mu_dict, logvar_dict = model.encode_modalities({modality: xb})

            if "modality" in latent:
                mu = mu_dict[modality]
                lv = logvar_dict[modality]
                z = mu if latent.endswith("_mean") else _sample_gaussian(mu, lv, gen)
            else:
                moe_out = model.mixture_of_experts(mu_dict, logvar_dict)
                if not isinstance(moe_out, (tuple, list)) or len(moe_out) < 2:
                    raise RuntimeError("mixture_of_experts must return (mu, logvar).")
                mu_z, logvar_z = moe_out[0], moe_out[1]
                z = mu_z if latent.endswith("_mean") else _sample_gaussian(mu_z, logvar_z, gen)

            zs.append(z.detach().cpu().numpy())

    return np.vstack(zs)


def cross_modal_predict(
    model,
    adata_src,
    src_mod: str,
    tgt_mod: str,
    device: str = "cpu",
    layer: Optional[str] = None,
    X_key: str = "X",
    batch_size: int = 512,
) -> np.ndarray:
    from .data import _get_matrix

    model.eval()
    X = _get_matrix(adata_src, layer=layer, X_key=X_key)
    if sp.issparse(X):
        X = X.toarray()

    preds = []
    with torch.no_grad():
        for start in range(0, X.shape[0], int(batch_size)):
            end = min(start + int(batch_size), X.shape[0])
            xb = torch.as_tensor(np.asarray(X[start:end]), dtype=torch.float32, device=device)

            mu_dict, logvar_dict = model.encode_modalities({src_mod: xb})
            mu_z, _ = model.mixture_of_experts(mu_dict, logvar_dict)
            xhat_dict = model.decode_modalities(mu_z)
            if tgt_mod not in xhat_dict:
                raise KeyError("Target modality %r not found. Available: %s" % (tgt_mod, list(xhat_dict.keys())))
            preds.append(xhat_dict[tgt_mod].detach().cpu().numpy())

    return np.vstack(preds) if preds else np.zeros((0, 0), dtype=float)


def denoise_adata(
    model,
    adata,
    modality: str,
    device: str = "cpu",
    layer: Optional[str] = None,
    X_key: str = "X",
    batch_size: int = 512,
    out_layer: Optional[str] = None,
    overwrite_X: bool = False,
    dtype: Optional[np.dtype] = np.float32,
) -> np.ndarray:
    X_hat = cross_modal_predict(
        model,
        adata_src=adata,
        src_mod=modality,
        tgt_mod=modality,
        device=device,
        layer=layer,
        X_key=X_key,
        batch_size=batch_size,
    )
    if dtype is not None:
        X_hat = np.asarray(X_hat, dtype=dtype)

    if overwrite_X:
        adata.X = X_hat
    elif out_layer is not None:
        adata.layers[out_layer] = X_hat

    return X_hat


def evaluate_alignment(
    Z1: Optional[np.ndarray] = None,
    Z2: Optional[np.ndarray] = None,
    model=None,
    adata1=None,
    adata2=None,
    mod1: Optional[str] = None,
    mod2: Optional[str] = None,
    device: str = "cpu",
    layer1: Optional[str] = None,
    layer2: Optional[str] = None,
    X_key1: str = "X",
    X_key2: str = "X",
    batch_size: int = 1024,
    latent: str = "moe_mean",
    latent1: Optional[str] = None,
    latent2: Optional[str] = None,
    random_state: int = 0,
    metric: str = "euclidean",
    k_mixing: int = 20,
    k_transfer: int = 15,
    modality_labels: Optional[np.ndarray] = None,
    labels_source: Optional[np.ndarray] = None,
    labels_target: Optional[np.ndarray] = None,
) -> Dict[str, Any]:
    out: Dict[str, Any] = {}

    lat1 = latent if latent1 is None else latent1
    lat2 = latent if latent2 is None else latent2

    if Z1 is None or Z2 is None:
        if model is None or adata1 is None or adata2 is None or mod1 is None or mod2 is None:
            raise ValueError("Provide either (Z1, Z2) or (model, adata1, adata2, mod1, mod2).")

        Z1 = encode_adata(model, adata1, modality=mod1, device=device, layer=layer1, X_key=X_key1,
                         batch_size=batch_size, latent=lat1, random_state=random_state)
        Z2 = encode_adata(model, adata2, modality=mod2, device=device, layer=layer2, X_key=X_key2,
                         batch_size=batch_size, latent=lat2, random_state=random_state)

    Z1 = np.asarray(Z1)
    Z2 = np.asarray(Z2)

    out["n1"] = int(Z1.shape[0])
    out["n2"] = int(Z2.shape[0])
    out["dim"] = int(Z1.shape[1]) if Z1.ndim == 2 else None
    out["latent1"] = lat1
    out["latent2"] = lat2

    if Z1.shape == Z2.shape and Z1.shape[0] > 1:
        out["foscttm"] = compute_foscttm(Z1, Z2, metric=metric)
    else:
        out["foscttm"] = None

    Z_concat = None
    if (Z1.ndim == 2 and Z2.ndim == 2 and Z1.shape[1] == Z2.shape[1]):
        Z_concat = np.vstack([Z1, Z2])
    if Z_concat is not None and Z_concat.shape[0] > 1:
        if modality_labels is None:
            modality_labels = np.concatenate([np.repeat("mod1", Z1.shape[0]), np.repeat("mod2", Z2.shape[0])])
        out["modality_mixing"] = compute_modality_mixing(Z_concat, modality_labels=np.asarray(modality_labels),
                                                         k=k_mixing, metric=metric)
    else:
        out["modality_mixing"] = None

    if labels_source is not None:
        if labels_target is not None:
            pred, acc, cm, order = label_transfer_knn(
                Z_source=Z1,
                labels_source=np.asarray(labels_source),
                Z_target=Z2,
                labels_target=np.asarray(labels_target),
                k=k_transfer,
                metric=metric,
                return_label_order=True,
            )
            out["label_transfer_label_order"] = order
        else:
            pred, acc, cm = label_transfer_knn(
                Z_source=Z1,
                labels_source=np.asarray(labels_source),
                Z_target=Z2,
                labels_target=None,
                k=k_transfer,
                metric=metric,
                return_label_order=False,
            )
            out["label_transfer_label_order"] = None

        out["label_transfer_pred"] = pred
        out["label_transfer_acc"] = acc
        out["label_transfer_cm"] = cm
    else:
        out["label_transfer_pred"] = None
        out["label_transfer_acc"] = None
        out["label_transfer_cm"] = None
        out["label_transfer_label_order"] = None

    return out

